from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union


class Agent(ABC):
    type: str

    @abstractmethod
    def build_prompt(self, messages: List[Dict[str, str]], tools: List[Dict[str, Any]]) -> List[Dict[str, str]]:
        r"""
        Formats messages for agent inference.

        Args:
            messages: the current messages.
            tools: the tool specification in the OpenAI format.

        Returns:
            messages: the formatted messages with tool information.
        """
        ...

    @abstractmethod
    def extract_tool(self, answer: str, tools: List[Dict[str, Any]]) -> Union[str, Tuple[str, str]]:
        r"""
        Extracts tool name and arguments from model outputs.

        Args:
            answer: the text to extract the tool from it.
            tools: the tool specification in the OpenAI format.

        Returns:
            response | (name, arguments): response text or tool name with JSON arguments if tool exists.
        """
        ...

    @abstractmethod
    def get_stop_word(self) -> Optional[str]:
        r"""
        Gets the stop word.

        Returns:
            stop_word (optional): the stop word.
        """
        ...



from typing import TYPE_CHECKING, Dict, List

from .aligned import Aligned
from .react import ReAct


if TYPE_CHECKING:
    from .types import Agent


_agents: Dict[str, "Agent"] = {}


def register_agent(agent: "Agent") -> None:
    _agents[agent.type] = agent


def list_agents() -> List[str]:
    return list(_agents.keys())


def get_agent(agent_type: str) -> "Agent":
    agent = _agents.get(agent_type, None)
    if agent is None:
        raise ValueError("Agent not found.")

    return agent


register_agent(Aligned())
register_agent(ReAct())



import os


def use_modelscope() -> bool:
    return os.environ.get("USE_MODELSCOPE_HUB", "0").lower() in ["true", "1"]


def try_download_model_from_ms(model_path: str) -> str:
    if not use_modelscope() or os.path.exists(model_path):
        return model_path

    try:
        from modelscope import snapshot_download

        return snapshot_download(model_path)
    except ImportError:
        raise ImportError("Please install modelscope via `pip install modelscope -U`")



from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, Any, AsyncGenerator, Dict, List, Optional, Tuple, Union

from transformers import AutoTokenizer, GenerationConfig
from typing_extensions import Self
from vllm import AsyncEngineArgs, AsyncLLMEngine, SamplingParams

from ..agent import get_agent, list_agents
from ..utils.modelscope import try_download_model_from_ms


if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace


@dataclass
class ChatConfig:
    r"""
    Creates configuration for a chat model.

    Methods:
        add_cli_args: adds arguments to a argument parser.
        from_cli_args: builds configuration based on the command line arguments.
    """

    name: str
    path: str
    device: List[int]
    port: int
    maxlen: int
    agent_type: str
    template: Optional[str]
    gen_config: Optional[str]

    @staticmethod
    def add_cli_args(parser: "ArgumentParser") -> None:
        parser.add_argument("--name", type=str, required=True, help="Model name.")
        parser.add_argument("--path", type=str, required=True, help="Model path or hub id.")
        parser.add_argument("--device", type=int, nargs="+", required=True, help="Device ids.")
        parser.add_argument("--port", type=int, required=True, help="API port.")
        parser.add_argument("--maxlen", type=int, default=2048, help="Max sequence length.")
        parser.add_argument("--agent_type", type=str, choices=list_agents(), default="react", help="Agent type.")
        parser.add_argument("--template", type=str, default=None, help="Template path.")
        parser.add_argument("--gen_config", type=str, default=None, help="Generation config path.")

    @classmethod
    def from_cli_args(cls, args: "Namespace") -> Self:
        attrs = [attr.name for attr in fields(cls)]
        return cls(**{attr: getattr(args, attr) for attr in attrs})


class ChatModel:
    r"""
    Creates a chat model for chat completions.

    Methods:
        chat: generates chat completions.
        stream_chat: streams chat completions.
        function_call: generates tool calls.
    """

    def __init__(self, config: "ChatConfig") -> None:
        config.path = try_download_model_from_ms(config.path)
        self.config = config
        self.name = config.name
        self._agent = get_agent(config.agent_type)
        self._init_vllm_engine()
        self._load_tokenizer()
        self._load_generation_config()

    def _init_vllm_engine(self) -> None:
        engine_args = AsyncEngineArgs(
            model=self.config.path,
            trust_remote_code=True,
            max_model_len=self.config.maxlen,
            tensor_parallel_size=len(self.config.device),
        )
        self._engine = AsyncLLMEngine.from_engine_args(engine_args)

    def _load_tokenizer(self) -> None:
        self._tokenizer = AutoTokenizer.from_pretrained(self.config.path, trust_remote_code=True)
        if self.config.template:
            with open(self.config.template, "r", encoding="utf-8") as f:
                self._tokenizer.chat_template = f.read()

        if self._tokenizer.chat_template is None:
            print("Chat template is not found, use the default one.")

    def _load_generation_config(self) -> None:
        try:
            generation_config_path = self.config.gen_config or self.config.path
            self._generation_config = GenerationConfig.from_pretrained(generation_config_path)
        except Exception:
            self._generation_config = GenerationConfig(
                pad_token_id=self._tokenizer.pad_token_id,
                bos_token_id=self._tokenizer.bos_token_id,
                eos_token_id=self._tokenizer.eos_token_id,
            )

        if not self._generation_config.temperature:
            self._generation_config.temperature = 1.0

        if not self._generation_config.top_p:
            self._generation_config.top_p = 1.0

        if not self._generation_config.max_new_tokens:
            self._generation_config.max_new_tokens = 2048

        if isinstance(self._generation_config.eos_token_id, int):
            self._generation_config.eos_token_id = [self._generation_config.eos_token_id]

        extra_special_tokens = []
        for eos_token_id in self._generation_config.eos_token_id:
            if eos_token_id != self._tokenizer.eos_token_id:
                extra_special_tokens.append(self._tokenizer.convert_ids_to_tokens(eos_token_id))

        self._engine.engine.tokenizer.tokenizer.add_special_tokens(
            {"additional_special_tokens": extra_special_tokens}, replace_additional_special_tokens=False
        )

    async def _generate(self, messages: List[Dict[str, str]], request_id: str, **gen_kwargs):
        input_ids = self._tokenizer.apply_chat_template(
            conversation=messages, tokenize=True, add_generation_prompt=True
        )
        sampling_params = SamplingParams(
            temperature=gen_kwargs.pop("temperature") or self._generation_config.temperature,
            top_p=gen_kwargs.pop("top_p") or self._generation_config.top_p,
            max_tokens=gen_kwargs.pop("max_tokens") or self._generation_config.max_new_tokens,
            stop=gen_kwargs.pop("stop", None),
            stop_token_ids=self._generation_config.eos_token_id + gen_kwargs.pop("stop_token_ids", []),
        )
        result_generator = self._engine.generate(
            inputs={"prompt_token_ids": input_ids}, sampling_params=sampling_params, request_id=request_id
        )
        return result_generator

    async def chat(self, messages: List[Dict[str, str]], request_id: str, **gen_kwargs) -> Tuple[str, int, int]:
        r"""
        Generates chat completions.

        Args:
            messages: input messages.
            request_id: request ID.
            temperature: generation parameter.
            top_p: generation parameter.
            max_tokens: generation parameter.
            stop: generation parameter.
            stop_token_ids: generation parameter.

        Returns:
            generated_text: the generated text.
            prompt_tokens: the number of prompt tokens.
            completion_tokens: the number of completion tokens.
        """
        final_result = None
        generator = await self._generate(messages, request_id, **gen_kwargs)
        async for result in generator:
            final_result = result

        generated_text = final_result.outputs[0].text
        prompt_tokens = len(final_result.prompt_token_ids)
        completion_tokens = len(final_result.outputs[0].token_ids)
        return generated_text, prompt_tokens, completion_tokens

    async def stream_chat(
        self, messages: List[Dict[str, str]], request_id: str, **gen_kwargs
    ) -> AsyncGenerator[str, None]:
        r"""
        Streams chat completions.

        Args:
            messages: input messages.
            request_id: request ID.
            temperature: generation parameter.
            top_p: generation parameter.
            max_tokens: generation parameter.
            stop: generation parameter.
            stop_token_ids: generation parameter.

        Returns:
            generated_token: the generated token.
        """
        generated_text = ""
        generator = await self._generate(messages, request_id, **gen_kwargs)
        async for result in generator:
            delta_text = result.outputs[0].text[len(generated_text) :]
            generated_text = result.outputs[0].text
            yield delta_text

    async def function_call(
        self, messages: List[Dict[str, str]], tools: List[Dict[str, Any]], request_id: str, **gen_kwargs
    ) -> Tuple[Union[str, Tuple[str, str]], int, int]:
        r"""
        Generates chat completions.

        Args:
            messages: input messages.
            tools: tools available.
            request_id: request ID.
            temperature: generation parameter.
            top_p: generation parameter.
            max_tokens: generation parameter.
            stop: generation parameter.
            stop_token_ids: generation parameter.

        Returns:
            response | (name, arguments): response text or tool name with JSON arguments if tool exists.
            prompt_tokens: the number of prompt tokens.
            completion_tokens: the number of completion tokens.
        """
        agent_messages = self._agent.build_prompt(messages, tools)
        stop_word = self._agent.get_stop_word()
        if stop_word is not None:
            stop_word_id = self._tokenizer.convert_tokens_to_ids(self._tokenizer.tokenize(stop_word)[0])
            gen_kwargs["stop_token_ids"] = gen_kwargs.pop("stop_token_ids", []) + [stop_word_id]

        generated_text, prompt_tokens, completion_tokens = await self.chat(agent_messages, request_id, **gen_kwargs)

        if stop_word is not None:
            stop_token = self._tokenizer.convert_ids_to_tokens(stop_word_id)
            if generated_text.endswith(stop_token):
                generated_text = generated_text[: -len(stop_token)]

        return self._agent.extract_tool(generated_text, tools), prompt_tokens, completion_tokens



from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Union

from typing_extensions import Literal


_PARAM_TYPE = {
    "string": str,
    "number": (int, float),
    "integer": int,
    "float": (int, float),
    "boolean": bool,
    "array": list,
    "object": dict,
}


@dataclass
class ParameterValue:
    type: Literal["string", "number", "integer", "float", "boolean", "array", "object"]
    description: str
    enum: List[str]
    item_type: str


@dataclass
class Parameter(ParameterValue):
    name: str


@dataclass
class Function:
    name: str
    description: str
    parameters: List[Parameter]
    required: List[str]

    def get_param_mapping(self) -> Dict[str, "ParameterValue"]:
        param_mapping = {}
        for param in self.parameters:
            param_dict = asdict(param)
            name = param_dict.pop("name")
            param_mapping[name] = ParameterValue(**param_dict)

        return param_mapping

    def verify_param(self, name: str, value: Any) -> bool:
        param_mapping = self.get_param_mapping()
        if name not in param_mapping:
            return False

        if not isinstance(value, _PARAM_TYPE[param_mapping[name].type]):
            return False

        if len(param_mapping[name].enum) and value not in param_mapping[name].enum:
            return False

        if param_mapping[name].item_type and not isinstance(value[0], _PARAM_TYPE[param_mapping[name].item_type]):
            return False

        return True


def get_functions(tools: List[Dict[str, Any]]) -> Dict[str, Function]:
    r"""
    Converts tools to functions.
    """
    functions = {}
    for tool in tools:
        if tool["type"] == "code_interpreter":
            functions["code_interpreter"] = Function(
                name="code_interpreter",
                description=(
                    "A Python code interpreter that executes Python code. "
                    "Enclose the code within triple backticks, e.g. ```code goes here```."
                ),
                parameters=[Parameter(name="code", type="string", description="Python code to execute.")],
                required=["code"],
            )
            break

        elif tool["type"] == "function":
            function_data: Dict[str, Union[str, Dict[str, Any]]] = tool["function"]
            properties: Dict[str, Dict[str, Any]] = function_data["parameters"].get("properties", {})
            parameters = [
                Parameter(
                    name=name,
                    type=prop.get("type", ""),
                    description=prop.get("description", ""),
                    enum=prop.get("enum", []),
                    item_type=prop.get("items", {}).get("type", ""),
                )
                for name, prop in properties.items()
            ]
            functions[function_data["name"]] = Function(
                name=function_data["name"],
                description=function_data["description"],
                parameters=parameters,
                required=function_data["parameters"].get("required", []),
            )

    return functions


if __name__ == "__main__":
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather in a given location",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"},
                        "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                    },
                    "required": ["location"],
                },
            },
        }
    ]
    functions = get_functions(tools)
    print(functions)
    print(functions["get_current_weather"].get_param_mapping())
    print(functions["get_current_weather"].verify_param("city", "San Francisco"))
    print(functions["get_current_weather"].verify_param("location", 0))
    print(functions["get_current_weather"].verify_param("location", "San Francisco"))
    print(functions["get_current_weather"].verify_param("unit", "test"))
    print(functions["get_current_weather"].verify_param("unit", "fahrenheit"))



import time
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from typing_extensions import Literal


class Role(str, Enum):
    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"
    FUNCTION = "function"
    TOOL = "tool"


class Finish(str, Enum):
    STOP = "stop"
    LENGTH = "length"
    TOOL = "tool_calls"


class ModelCard(BaseModel):
    id: str
    object: Literal["model"] = "model"
    created: int = Field(default_factory=lambda: int(time.time()))
    owned_by: Literal["owner"] = "owner"


class ModelList(BaseModel):
    object: Literal["list"] = "list"
    data: List[ModelCard] = []


class Function(BaseModel):
    name: str
    arguments: str


class FunctionDefinition(BaseModel):
    name: str
    description: str
    parameters: Dict[str, Any]


class FunctionCall(BaseModel):
    id: str
    type: Literal["function"] = "function"
    function: Function


class FunctionAvailable(BaseModel):
    type: Literal["function", "code_interpreter"] = "function"
    function: Optional[FunctionDefinition] = None


class ChatMessage(BaseModel):
    role: Role
    content: str


class ChatCompletionMessage(BaseModel):
    role: Optional[Role] = None
    content: Optional[str] = None
    tool_calls: Optional[List[FunctionCall]] = None


class UsageInfo(BaseModel):
    prompt_tokens: int
    completion_tokens: Optional[int] = None
    total_tokens: int


class ChatCompletionRequest(BaseModel):
    model: str
    messages: List[ChatMessage]
    tools: Optional[List[FunctionAvailable]] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    n: int = 1
    max_tokens: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    stream: Optional[bool] = False


class ChatCompletionResponseChoice(BaseModel):
    index: int
    message: ChatCompletionMessage
    finish_reason: Finish


class ChatCompletionStreamResponseChoice(BaseModel):
    index: int
    delta: ChatCompletionMessage
    finish_reason: Optional[Finish] = None


class ChatCompletionResponse(BaseModel):
    id: str
    object: Literal["chat.completion"] = "chat.completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseChoice]
    usage: UsageInfo


class ChatCompletionStreamResponse(BaseModel):
    id: str
    object: Literal["chat.completion.chunk"] = "chat.completion.chunk"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionStreamResponseChoice]


class EmbeddingsRequest(BaseModel):
    input: Union[str, List[str]]
    model: str
    encoding_format: Literal["base64", "float"] = "float"


class Embeddings(BaseModel):
    object: Literal["embedding"] = "embedding"
    embedding: Union[List[float], bytes]
    index: int


class EmbeddingsResponse(BaseModel):
    object: Literal["list"] = "list"
    data: List[Embeddings]
    model: str
    usage: UsageInfo



import argparse
import base64
from contextlib import asynccontextmanager

import uvicorn
from fastapi import FastAPI, status

from ..model import EmbedConfig, EmbedModel
from .protocol import (
    Embeddings,
    EmbeddingsRequest,
    EmbeddingsResponse,
    UsageInfo,
)


async def _create_local_embeddings(request: "EmbeddingsRequest", model: "EmbedModel") -> "EmbeddingsResponse":
    texts = request.input
    if isinstance(texts, str):
        texts = [texts]

    embed_output, embed_tokens = await model.embed(texts)
    embeddings = []
    for i, embed_data in enumerate(embed_output):
        if request.encoding_format == "base64":
            embedding = base64.b64encode(embed_data)
        else:
            embedding = embed_data.tolist()

        embeddings.append(Embeddings(embedding=embedding, index=i))

    return EmbeddingsResponse(
        data=embeddings,
        model=request.model,
        usage=UsageInfo(prompt_tokens=embed_tokens, completion_tokens=None, total_tokens=embed_tokens),
    )


def launch_server(config: "EmbedConfig") -> None:
    model = EmbedModel(config)

    @asynccontextmanager
    async def lifespan(app: "FastAPI"):
        await model.startup()
        yield
        await model.shutdown()

    app = FastAPI(lifespan=lifespan)

    @app.post("/v1/embeddings", response_model=EmbeddingsResponse, status_code=status.HTTP_200_OK)
    async def create_embeddings_v1(request: "EmbeddingsRequest"):
        return await _create_local_embeddings(request, model)

    uvicorn.run(app, port=config.port)


def main():
    parser = argparse.ArgumentParser()
    EmbedConfig.add_cli_args(parser)
    args = parser.parse_args()
    config = EmbedConfig.from_cli_args(args)
    launch_server(config)


if __name__ == "__main__":
    main()



from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, List, Tuple

from infinity_emb import AsyncEmbeddingEngine, EngineArgs
from typing_extensions import Self

from ..utils.modelscope import try_download_model_from_ms


if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace

    from numpy import float32
    from numpy.typing import NDArray


@dataclass
class EmbedConfig:
    r"""
    Creates configuration for an embedding model.

    Methods:
        add_cli_args: adds arguments to a argument parser.
        from_cli_args: builds configuration based on the command line arguments.
    """

    name: str
    path: str
    device: List[int]
    port: int
    batch_size: int

    @staticmethod
    def add_cli_args(parser: "ArgumentParser") -> None:
        parser.add_argument("--name", type=str, required=True, help="Model name.")
        parser.add_argument("--path", type=str, required=True, help="Model path or hub id.")
        parser.add_argument("--device", type=int, nargs="+", required=True, help="Device ids.")
        parser.add_argument("--port", type=int, required=True, help="API port.")
        parser.add_argument("--batch_size", type=int, default=64, help="Batch size.")

    @classmethod
    def from_cli_args(cls, args: "Namespace") -> Self:
        attrs = [attr.name for attr in fields(cls)]
        return cls(**{attr: getattr(args, attr) for attr in attrs})


class EmbedModel:
    r"""
    Creates an embedding model for text embeddings.

    Methods:
        startup: starts the embedding engine.
        shutdown: stops the embedding engine.
        embed: calculates text embeddings.
    """

    def __init__(self, config: "EmbedConfig") -> None:
        config.path = try_download_model_from_ms(config.path)
        self.config = config
        self.name = config.name
        self._init_infinity_engine()

    def _init_infinity_engine(self) -> None:
        if len(self.config.device) != 1:
            raise ValueError("Embedding model does not support multi-GPUs yet.")

        engine_args = EngineArgs(
            model_name_or_path=self.config.path,
            batch_size=self.config.batch_size,
        )
        self._engine = AsyncEmbeddingEngine.from_args(engine_args)

    async def startup(self) -> None:
        r"""
        Starts the embedding engine.
        """
        await self._engine.astart()

    async def shutdown(self) -> None:
        r"""
        Stops the embedding engine.
        """
        await self._engine.astop()

    async def embed(self, texts: List[str]) -> Tuple[List["NDArray[float32]"], int]:
        r"""
        Calculates the text embeddings.

        Args:
            texts: the batched text input.

        Returns:
            embeddings: the batched embeddings.
            usage: the number of input tokens.
        """
        return await self._engine.embed(texts)



import argparse
import uuid
from typing import Any, AsyncGenerator, Dict

import uvicorn
from fastapi import FastAPI, HTTPException, status
from sse_starlette import EventSourceResponse

from ..model import ChatConfig, ChatModel
from ..utils.generic import dictify
from .common import create_stream_chunk
from .protocol import (
    ChatCompletionMessage,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    Finish,
    Function,
    FunctionCall,
    Role,
    UsageInfo,
)


async def _create_stream_chat_completion(
    request: "ChatCompletionRequest", model: "ChatModel", input_kwargs: Dict[str, Any]
) -> AsyncGenerator[str, None]:
    yield create_stream_chunk(
        input_kwargs["request_id"], request.model, ChatCompletionMessage(role=Role.ASSISTANT, content="")
    )
    async for new_token in model.stream_chat(**input_kwargs):
        yield create_stream_chunk(input_kwargs["request_id"], request.model, ChatCompletionMessage(content=new_token))

    yield create_stream_chunk(
        input_kwargs["request_id"], request.model, ChatCompletionMessage(), finish_reason=Finish.STOP
    )
    yield "[DONE]"


async def _create_local_chat_completion(
    request: "ChatCompletionRequest", model: "ChatModel"
) -> "ChatCompletionResponse":
    msg_id = uuid.uuid4().hex
    input_kwargs = {
        "messages": [dictify(message) for message in request.messages],
        "request_id": "chatcmpl-{}".format(msg_id),
        "temperature": request.temperature,
        "top_p": request.top_p,
        "max_tokens": request.max_tokens,
        "stop": request.stop,
    }

    if request.n != 1:
        raise NotImplementedError("Multiple responses are not supported yet.")

    if request.stream:
        if request.tools is not None:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Cannot stream function calls.")

        generator = _create_stream_chat_completion(request, model, input_kwargs)
        return EventSourceResponse(generator, media_type="text/event-stream")

    if request.tools is not None:
        input_kwargs["tools"] = [dictify(tool) for tool in request.tools]
        result, prompt_tokens, completion_tokens = await model.function_call(**input_kwargs)
    else:
        result, prompt_tokens, completion_tokens = await model.chat(**input_kwargs)

    if isinstance(result, tuple):
        name, arguments = result[0], result[1]
        tool_call = FunctionCall(id="call_{}".format(msg_id), function=Function(name=name, arguments=arguments))
        choice = ChatCompletionResponseChoice(
            index=0,
            message=ChatCompletionMessage(role=Role.ASSISTANT, tool_calls=[tool_call]),
            finish_reason=Finish.TOOL,
        )
    else:
        choice = ChatCompletionResponseChoice(
            index=0, message=ChatCompletionMessage(role=Role.ASSISTANT, content=result), finish_reason=Finish.STOP
        )

    usage = UsageInfo(
        prompt_tokens=prompt_tokens,
        completion_tokens=completion_tokens,
        total_tokens=prompt_tokens + completion_tokens,
    )
    return ChatCompletionResponse(
        id=input_kwargs["request_id"],
        model=request.model,
        choices=[choice],
        usage=usage,
    )


def launch_server(config: "ChatConfig") -> None:
    model = ChatModel(config)
    app = FastAPI()

    @app.post("/v1/chat/completions", response_model=ChatCompletionResponse, status_code=status.HTTP_200_OK)
    async def create_chat_completion(request: "ChatCompletionRequest"):
        return await _create_local_chat_completion(request, model)

    uvicorn.run(app, port=config.port)


def main():
    parser = argparse.ArgumentParser()
    ChatConfig.add_cli_args(parser)
    args = parser.parse_args()
    config = ChatConfig.from_cli_args(args)
    launch_server(config)


if __name__ == "__main__":
    main()



import json
import re
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union

from .function import get_functions
from .types import Agent


if TYPE_CHECKING:
    from .function import Function


JSON_FORMAT_PROMPT = (
    """, in a JSON format representing the kwargs (e.g. ```{"input": "hello world", "num_beams": 5}```)"""
)


CODE_FORMAT_PROMPT = ", in triple backticks (e.g. ```code goes here```)"


REACT_PROMPT = (
    "Answer the following questions as best you can. "
    "You have access to the following tools:\n{tool_text}"
    "Use the following format to answer the question:\n"
    "```\n"
    "Thought: I need to use a tool to help me answer the question.\n"
    "Action: the action to take, should be one of [{tool_names}] if using a tool.\n"
    "Action Input: the input to the action{format_prompt}.\n"
    "```\n\n"
    "Please ALWAYS start with a Thought. "
    "If this format is used, the user will respond in the following format:\n"
    "```\n"
    "Observation: [tool response]\n"
    "```\n\n"
    "You should keep repeating the above format until you have enough information "
    "to answer the question without using any more tools. At that point, you MUST respond "
    "in the one of the following two formats:\n"
    "```\n"
    "Thought: I can answer without using any more tools.\n"
    "Answer: [your answer here]\n"
    "```\n"
    "```\n"
    "Thought: I cannot answer the question with the provided tools.\n"
    "Answer: Sorry, I cannot answer your question.\n"
    "```\n\n"
    "Below is the current conversation consisting of interleaving human and assistant messages.\n"
    "```\n"
    "{messages_text}"
    "```"
)


class ReAct(Agent):
    def __init__(self) -> None:
        self.type = "react"

    def _build_tool_text(self, functions: Dict[str, "Function"]) -> str:
        tool_text = ""
        for func in functions.values():
            param_text = ""
            for param in func.parameters:
                required = ", required" if param.name in func.required else ""
                enum = ", should be one of [{}]".format(", ".join(param.enum)) if len(param.enum) else ""
                item_type = ", where each item should be {}".format(param.item_type) if param.item_type else ""
                param_text += "  - {name} ({type}{required}): {desc}{enum}{item_type}\n".format(
                    name=param.name,
                    type=param.type,
                    required=required,
                    desc=param.description,
                    enum=enum,
                    item_type=item_type,
                )

            tool_text += "> Tool Name: {name}\nTool Description: {desc}\nTool Args:\n{args}\n".format(
                name=func.name, desc=func.description, args=param_text
            )

        return tool_text

    def _build_messages_text(self, messages: List[Dict[str, str]]) -> str:
        messages_text = ""
        for message in messages:
            if message["role"] == "user":
                messages_text += "Human: {}\n".format(message["content"])

            elif message["role"] == "assistant":
                messages_text += "Assistant: {}\n".format(message["content"])

            elif message["role"] == "tool":
                messages_text += "Observation: {}\n".format(message["content"])

            elif message["role"] == "function":
                try:
                    tool_call: Dict[str, str] = json.loads(message["content"])
                except json.JSONDecodeError:
                    print("Invalid json string.")

                messages_text += "Action: {}\nAction Input: {}\n".format(
                    tool_call.get("name", ""), tool_call.get("arguments", "")
                )

        return messages_text

    def build_prompt(self, messages: List[Dict[str, str]], tools: List[Dict[str, Any]]) -> List[Dict[str, str]]:
        functions = get_functions(tools)
        prompt = REACT_PROMPT.format(
            tool_text=self._build_tool_text(functions),
            tool_names=", ".join(functions.keys()),
            format_prompt=CODE_FORMAT_PROMPT if "code_interpreter" in functions else JSON_FORMAT_PROMPT,
            messages_text=self._build_messages_text(messages),
        )
        return [{"role": "user", "content": prompt}]

    def extract_tool(self, answer: str, tools: Dict[str, Any]) -> Union[str, Tuple[str, str]]:
        functions = get_functions(tools)
        regex = re.compile(r"Action:\s*([a-zA-Z0-9_]+)\s*Action Input:\s*(.*)", re.DOTALL)
        action_match = re.search(regex, answer)
        if action_match:
            tool_name = action_match.group(1).strip()
            if tool_name not in functions:
                return "Failed: tool does not exist."

            tool_input = action_match.group(2).strip().strip('"')
            if "code_interpreter" in functions:
                arguments = {"code": tool_input}
            else:
                try:
                    arguments = json.loads(tool_input.strip("```"))
                except json.JSONDecodeError:
                    return "Failed: params not found."

            for name, value in arguments.items():
                if not functions[tool_name].verify_param(name, value):
                    return "Failed: invalid params."

            return tool_name, json.dumps(arguments, ensure_ascii=False)
        else:
            regex = re.compile(r"Answer:\s*(.*)", re.DOTALL)
            answer_match = re.search(regex, answer)
            if answer_match:
                return answer_match.group(1).strip()
            else:
                return "Failed: answers not found."

    def get_stop_word(self) -> str:
        return "Observation"


if __name__ == "__main__":
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather in a given location",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"},
                        "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                    },
                    "required": ["location"],
                },
            },
        }
    ]
    react = ReAct()
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    print(react.build_prompt(messages, tools))

    answer = """Action: get_current_weather\nAction Input: {\n"location": "Boston, MA"\n}"""
    print(react.extract_tool(answer, tools))

    answer = "Thought: I can answer.\nAnswer: The weather in Boston is currently sunny."
    print(react.extract_tool(answer, tools))



import argparse
from enum import Enum, unique
from typing import Any, Dict

import click
import yaml
from openai import OpenAI


try:
    import platform

    if platform.system() != "Windows":
        import readline  # noqa: F401
except ImportError:
    print("Install `readline` for a better experience.")


@unique
class Action(str, Enum):
    CHAT = "chat"
    EMBED = "embed"
    TOOL = "tool"


def test_chat(client: "OpenAI", query: str) -> None:
    stream = client.chat.completions.create(
        messages=[{"role": "user", "content": query}],
        model="gpt-3.5-turbo",
        stream=True,
    )
    print("Assistant: ", end="")
    for chunk in filter(lambda p: p.choices[0].delta.content is not None, stream):
        print(chunk.choices[0].delta.content, end="", flush=True)
    print()


def test_embed(client: "OpenAI", query: str) -> None:
    data = client.embeddings.create(input=query, model="text-embedding-ada-002", encoding_format="float").data
    for embedding in data:
        print(embedding.embedding)


def test_tool(client: "OpenAI") -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather in a given location",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"}
                    },
                    "required": ["location"],
                },
            },
        }
    ]
    result = client.chat.completions.create(
        messages=[{"role": "user", "content": "What is the weather like in Boston?"}],
        model="gpt-3.5-turbo",
        tools=tools,
    )
    print(result.choices[0].message)
    result = client.chat.completions.create(
        messages=[
            {"role": "user", "content": "What is the weather like in Boston?"},
            {
                "role": "function",
                "content": """{"name": "get_current_weather", "arguments": {"location": "Boston, MA"}}""",
            },
            {"role": "tool", "content": """{"temperature": 22, "unit": "celsius", "description": "Sunny"}"""},
        ],
        model="gpt-3.5-turbo",
        tools=tools,
    )
    print(result.choices[0].message)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", type=str, required=True, help="Path to config file.")
    args = parser.parse_args()
    with open(getattr(args, "config"), "r", encoding="utf-8") as f:
        config: Dict[str, Dict[str, Any]] = yaml.safe_load(f)

    client = OpenAI(
        api_key=config["service"].get("api_key", "0"),
        base_url="http://{host}:{port}/v1".format(
            host=config["service"].get("host", "localhost"),
            port=config["service"].get("port", 8000),
        ),
    )
    action = click.prompt("Action", type=click.Choice([act.value for act in Action]))
    if action == Action.TOOL:
        test_tool(client)
    else:
        while True:
            query = click.prompt("User", type=str)
            if query == "exit":
                break

            if action == Action.CHAT:
                test_chat(client, query)
            elif action == Action.EMBED:
                test_embed(client, query)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING, Optional

from ..utils.generic import jsonify
from .protocol import ChatCompletionStreamResponse, ChatCompletionStreamResponseChoice


if TYPE_CHECKING:
    from subprocess import Popen

    from .protocol import ChatCompletionMessage, Finish


def print_subprocess_stdout(process: "Popen") -> None:
    while process.stdout.readable():
        line: bytes = process.stdout.readline()

        if not line:
            break

        print(line.decode("utf-8").strip())


def create_stream_chunk(
    request_id: str,
    model: str,
    delta: "ChatCompletionMessage",
    index: Optional[int] = 0,
    finish_reason: Optional["Finish"] = None,
) -> str:
    choice = ChatCompletionStreamResponseChoice(index=index, delta=delta, finish_reason=finish_reason)
    chunk = ChatCompletionStreamResponse(id=request_id, model=model, choices=[choice])
    return jsonify(chunk)



import os
import re

from setuptools import find_packages, setup


def get_version():
    with open(os.path.join("src", "imitater", "__init__.py"), "r", encoding="utf-8") as f:
        file_content = f.read()
        pattern = r"{0}\W*=\W*\"([^\"]+)\"".format("__version__")
        version, = re.findall(pattern, file_content)
        return version


def get_requires():
    with open("requirements.txt", "r", encoding="utf-8") as f:
        file_content = f.read()
        lines = [line.strip() for line in file_content.strip().split("\n") if not line.startswith("#")]
        return lines


def main():

    setup(
        name="imitater",
        version=get_version(),
        author="the-seeds",
        description="",
        long_description=open("README.md", "r", encoding="utf-8").read(),
        long_description_content_type="text/markdown",
        keywords=["LLM"],
        license="Apache 2.0 License",
        url="https://github.com/the-seeds/imitater",
        package_dir={"": "src"},
        packages=find_packages("src"),
        python_requires=">=3.9.0",
        install_requires=get_requires(),
        entry_points={"console_scripts": ["imitater = imitater.service.app:main"]},
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
        ]
    )


if __name__ == "__main__":
    main()



import json
from typing import TYPE_CHECKING, Any, Dict


if TYPE_CHECKING:
    from pydantic import BaseModel


def dictify(data: "BaseModel") -> Dict[str, Any]:
    try:  # pydantic v2
        return data.model_dump(exclude_unset=True)
    except AttributeError:  # pydantic v1
        return data.dict(exclude_unset=True)


def jsonify(data: "BaseModel") -> str:
    try:  # pydantic v2
        return json.dumps(data.model_dump(exclude_unset=True), ensure_ascii=False)
    except AttributeError:  # pydantic v1
        return data.json(exclude_unset=True, ensure_ascii=False)



import argparse
import os
from copy import deepcopy
from subprocess import PIPE, STDOUT, Popen
from threading import Thread
from typing import Annotated, Any, AsyncGenerator, Dict, List, Optional, Union

import uvicorn
import yaml
from fastapi import Depends, FastAPI, HTTPException, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.security.http import HTTPAuthorizationCredentials, HTTPBearer
from openai import AsyncOpenAI, AsyncStream
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from sse_starlette import EventSourceResponse

from ..utils.generic import dictify
from .common import create_stream_chunk, print_subprocess_stdout
from .protocol import (
    ChatCompletionMessage,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    Embeddings,
    EmbeddingsRequest,
    EmbeddingsResponse,
    Finish,
    Function,
    FunctionCall,
    ModelCard,
    ModelList,
    Role,
    UsageInfo,
)


async def _stream_openai_chat_completion(response: AsyncStream[ChatCompletionChunk]) -> AsyncGenerator[str, None]:
    request_id, model = None, None
    async for chunk in response:
        if request_id is None:
            request_id = chunk.id
            model = chunk.model
            yield create_stream_chunk(request_id, model, ChatCompletionMessage(role=Role.ASSISTANT, content=""))

        new_token = chunk.choices[0].delta.content
        if new_token:
            yield create_stream_chunk(request_id, model, ChatCompletionMessage(content=new_token))

    yield create_stream_chunk(request_id, model, ChatCompletionMessage(), finish_reason=Finish.STOP)
    yield "[DONE]"


async def _create_openai_chat_completion(
    request: "ChatCompletionRequest", model: "AsyncOpenAI"
) -> "ChatCompletionResponse":
    response: "ChatCompletion" = await model.chat.completions.create(
        model=request.model,
        messages=[dictify(message) for message in request.messages],
        tools=[dictify(tool) for tool in request.tools] if request.tools is not None else None,
        temperature=request.temperature,
        top_p=request.top_p,
        n=request.n,
        max_tokens=request.max_tokens,
        stream=request.stream,
    )

    if request.stream:
        return EventSourceResponse(_stream_openai_chat_completion(response), media_type="text/event-stream")

    choices = []
    for i, choice in enumerate(response.choices):
        if choice.message.tool_calls is not None:
            tool_calls = []
            for tool_call in choice.message.tool_calls:
                name = tool_call.function.name
                arguments = tool_call.function.arguments
                tool_calls.append(FunctionCall(id=tool_call.id, function=Function(name=name, arguments=arguments)))

            choices.append(
                ChatCompletionResponseChoice(
                    index=i,
                    message=ChatCompletionMessage(role=Role.ASSISTANT, tool_calls=tool_calls),
                    finish_reason=Finish.TOOL,
                )
            )
        else:
            content = choice.message.content
            choices.append(
                ChatCompletionResponseChoice(
                    index=i,
                    message=ChatCompletionMessage(role=Role.ASSISTANT, content=content),
                    finish_reason=Finish.STOP,
                )
            )

    usage = UsageInfo(
        prompt_tokens=response.usage.prompt_tokens,
        completion_tokens=response.usage.completion_tokens,
        total_tokens=response.usage.total_tokens,
    )
    return ChatCompletionResponse(
        id=response.id,
        model=response.model,
        choices=choices,
        usage=usage,
    )


async def _create_openai_embeddings(request: "EmbeddingsRequest", model: "AsyncOpenAI") -> "EmbeddingsResponse":
    response = await model.embeddings.create(
        model=request.model,
        input=request.input,
        encoding_format=request.encoding_format,
    )

    embeddings = []
    for i, embedding in enumerate(response.data):
        embeddings.append(Embeddings(embedding=embedding.embedding, index=i))

    usage = UsageInfo(
        prompt_tokens=response.usage.prompt_tokens,
        total_tokens=response.usage.total_tokens,
    )
    return EmbeddingsResponse(
        data=embeddings,
        model=response.model,
        usage=usage,
    )


def _launch_chat_server(chat_config: Dict[str, Any]) -> "Popen":
    cmd = "python -m imitater.service.chat"
    cmd += " --name {}".format(chat_config["name"])
    cmd += " --path {}".format(chat_config["path"])
    cmd += " --device {}".format(" ".join(map(str, chat_config["device"])))
    cmd += " --port {}".format(chat_config["port"])
    cmd += " --maxlen {}".format(chat_config["maxlen"]) if "maxlen" in chat_config else ""
    cmd += " --agent_type {}".format(chat_config["agent_type"]) if "agent_type" in chat_config else ""
    cmd += " --template {}".format(chat_config["template"]) if "template" in chat_config else ""
    cmd += " --gen_config {}".format(chat_config["gen_config"]) if "gen_config" in chat_config else ""
    env = deepcopy(os.environ)
    env["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, chat_config["device"]))
    return Popen(cmd, env=env, shell=True, stdout=PIPE, stderr=STDOUT)


def _launch_embed_server(embed_config: Dict[str, Any]) -> "Popen":
    cmd = "python -m imitater.service.embed"
    cmd += " --name {}".format(embed_config["name"])
    cmd += " --path {}".format(embed_config["path"])
    cmd += " --device {}".format(" ".join(map(str, embed_config["device"])))
    cmd += " --port {}".format(embed_config["port"])
    cmd += " --batch_size {}".format(embed_config["batch_size"]) if "batch_size" in embed_config else ""
    env = deepcopy(os.environ)
    env["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, embed_config["device"]))
    return Popen(cmd, env=env, shell=True, stdout=PIPE, stderr=STDOUT)


def launch_server(config_file: str) -> None:
    with open(config_file, "r", encoding="utf-8") as f:
        config: Dict[str, Union[Dict[str, Any], List[Dict[str, Any]]]] = yaml.safe_load(f)

    port = config["service"].get("port", 8000)
    api_key = config["service"].get("api_key", "")
    chat_models: Dict[str, "AsyncOpenAI"] = {}
    embed_models: Dict[str, "AsyncOpenAI"] = {}
    processes: List["Popen"] = []

    if "chat" in config:
        for chat_config in config["chat"]:
            if "token" in chat_config:
                chat_models[chat_config["name"]] = AsyncOpenAI(api_key=chat_config["token"])
            else:
                processes.append(_launch_chat_server(chat_config))
                chat_models[chat_config["name"]] = AsyncOpenAI(
                    api_key="0", base_url="http://localhost:{}/v1".format(chat_config["port"])
                )

    if "embed" in config:
        for embed_config in config["embed"]:
            if "token" in embed_config:
                embed_models[embed_config["name"]] = AsyncOpenAI(api_key=embed_config["token"])
            else:
                processes.append(_launch_embed_server(embed_config))
                embed_models[embed_config["name"]] = AsyncOpenAI(
                    api_key="0", base_url="http://localhost:{}/v1".format(embed_config["port"])
                )

    app = FastAPI()
    app.add_middleware(
        CORSMiddleware, allow_origins=["*"], allow_credentials=True, allow_methods=["*"], allow_headers=["*"]
    )

    security = HTTPBearer(auto_error=False)

    async def verify_api_key(auth: Annotated[Optional[HTTPAuthorizationCredentials], Depends(security)]) -> None:
        if api_key and (auth is None or auth.credentials != api_key):
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid API key.")

    @app.get("/v1/models", response_model=ModelList, dependencies=[Depends(verify_api_key)])
    async def list_models():
        model_names = set()
        model_names.update(chat_models.keys())
        model_names.update(embed_models.keys())
        return ModelList(data=[ModelCard(id=name) for name in model_names])

    @app.post(
        "/v1/chat/completions",
        response_model=ChatCompletionResponse,
        status_code=status.HTTP_200_OK,
        dependencies=[Depends(verify_api_key)],
    )
    async def create_chat_completion(request: "ChatCompletionRequest"):
        if request.model not in chat_models:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Model not found.")
        return await _create_openai_chat_completion(request, chat_models[request.model])

    @app.post(
        "/v1/embeddings",
        response_model=EmbeddingsResponse,
        status_code=status.HTTP_200_OK,
        dependencies=[Depends(verify_api_key)],
    )
    async def create_embeddings(request: "EmbeddingsRequest"):
        if request.model not in embed_models:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Model not found.")
        return await _create_openai_embeddings(request, embed_models[request.model])

    for process in processes:
        thread = Thread(target=print_subprocess_stdout, args=[process], daemon=True)
        thread.start()

    uvicorn.run(app, host="0.0.0.0", port=port)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", type=str, required=True, help="Path to config file.")
    args = parser.parse_args()
    launch_server(getattr(args, "config"))


if __name__ == "__main__":
    main()
